# Copyright 2018      Cyril Roelandt
#
# Licensed under the 3-clause BSD license. See the LICENSE file.
import requests
import upt


class CPANPackage(upt.Package):
    pass


class CPANFrontend(upt.Frontend):
    name = 'cpan'

    def _get_homepage(self, json_resources):
        homepage = json_resources.get('homepage')
        if homepage is None:
            for kind in ('bugtracker', 'repository'):
                d = json_resources.get(kind, {})
                homepage = d.get('web')
                if homepage is not None:
                    break
        return homepage

    def _get_licenses(self, licenses):
        # List taken from https://metacpan.org/pod/CPAN::Meta::Spec#license .
        # Not included here:
        # - gpl_1: deprecated for ages
        # - ssleay: not even listed on https://spdx.org/licenses/
        # - open_source: not a real license
        # - restricte: not a real license
        # - unrestricte: not a real license
        # - unknown: not a real license
        metacpan_licenses = {
            'agpl_3':
                upt.licenses.GNUAfferoGeneralPublicLicenseThreeDotZeroPlus,
            'apache_1_1': upt.licenses.ApacheLicenseOneDotOne,
            'apache_2_0': upt.licenses.ApacheLicenseTwoDotZero,
            'artistic_1': upt.licenses.ArtisticLicenseOneDotZero,
            'artistic_2': upt.licenses.ApacheLicenseTwoDotZero,
            'bsd': upt.licenses.BSDThreeClauseLicense,
            'freebsd': upt.licenses.BSDTwoClauseLicense,
            'gfdl_1_2': upt.licenses.GNUFreeDocumentationLicenseOneDotTwo,
            'gfdl_1_3': upt.licenses.GNUFreeDocumentationLicenseOneDotThree,
            'gpl_2': upt.licenses.GNUGeneralPublicLicenseTwo,
            'gpl_3': upt.licenses.GNUGeneralPublicLicenseThree,
            'lgpl_2_1': upt.licenses.GNULesserGeneralPublicLicenseTwoDotOne,
            'lgpl_3_0': upt.licenses.GNULesserGeneralPublicLicenseThreeDotZero,
            'mit': upt.licenses.MITLicense,
            'mozilla_1_0': upt.licenses.MozillaPublicLicenseOneDotZero,
            'mozilla_1_1': upt.licenses.MozillaPublicLicenseOneDotOne,
            'openssl': upt.licenses.OpenSSLLicense,
            'perl_5': upt.licenses.PerlLicense,
            'qpl_1_0': upt.licenses.QPublicLicenseOneDotZero,
            'sun': upt.licenses.SunIndustryStandardsSourceLicenceOneDotOne,
            'zlib': upt.licenses.ZlibLicense,
        }

        return [metacpan_licenses.get(l, upt.licenses.UnknownLicense)()
                for l in licenses]

    def _get_requirements(self, json_dependencies):
        phases_to_upt_phases = {
            'configure': 'config',
            'build': 'build',
            'runtime': 'run',
            'test': 'test',
        }

        requirements = {}
        for phase, dependencies in json_dependencies.items():
            # We do not really care about this phase.
            if phase == 'develop':
                continue

            upt_phase = phases_to_upt_phases.get(phase, None)
            if upt_phase is None:
                continue

            requirements[upt_phase] = [
                upt.PackageRequirement(module,
                                       '' if spec == '0' else f'>={spec}')
                for (module, spec) in dependencies.get('requires', {}).items()
                if module != 'perl'  # This seems implicit
            ]

        return requirements

    def _get_archives(self, json):
        try:
            url = json['download_url']
            stat = json.get('stat', {})
            return [
                upt.Archive(url=url, size=stat.get('size', 0))
            ]
        except KeyError:
            return []

    def _get_json_data(self, pkg_name, version=None):
        # The following document https://metacpan.org/pod/CPAN::Meta::Spec
        # is quite useful to understand the JSON we get.
        release_name = pkg_name.replace('::', '-')
        url = f'https://fastapi.metacpan.org/v1/release/{release_name}'
        r = requests.get(url)
        if not r.ok:
            raise upt.InvalidPackageNameError(self.name, pkg_name)
        json = r.json()

        if version is not None:
            url = 'https://fastapi.metacpan.org/v1/release/'
            url += f'{json["author"]}/{release_name}-{version}'
            r = requests.get(url)
            if not r.ok:
                raise upt.InvalidPackageNameError(self.name, pkg_name)
            return r.json()['release']
        else:
            return json

    def parse(self, pkg_name, version=None):
        json = self._get_json_data(pkg_name, version)
        metadata = json.get('metadata', {})
        version = metadata.get('version', '')
        homepage = self._get_homepage(json.get('resources', {}))

        pkg_args = {
            'summary': json.get('abstract', ''),
            'homepage': homepage,
            'licenses': self._get_licenses(metadata.get('license', [])),
            'archives': self._get_archives(json),
            'requirements': self._get_requirements(metadata.get('prereqs', {}))
        }
        return CPANPackage(pkg_name, version, **pkg_args)



# Copyright 2018      Cyril Roelandt
#
# Licensed under the 3-clause BSD license. See the LICENSE file.
import unittest

import requests_mock

import upt
from upt_cpan.upt_cpan import CPANFrontend


class TestCPANFrontend(unittest.TestCase):
    def setUp(self):
        self.frontend = CPANFrontend()

    def test_invalid_package_name(self):
        with self.assertRaises(upt.InvalidPackageNameError):
            self.frontend.parse('probably-invalid-package-name')

    def test_get_homepage(self):
        out = self.frontend._get_homepage({})
        self.assertIsNone(out)

        out = self.frontend._get_homepage({'homepage': 'foo'})
        self.assertEqual(out, 'foo')

        out = self.frontend._get_homepage({'bugtracker': {'web': 'foo'}})
        self.assertEqual(out, 'foo')

        out = self.frontend._get_homepage({'repository': {'web': 'foo'}})
        self.assertEqual(out, 'foo')

        json = {'repository': {'web': 'bar'}, 'homepage': 'foo'}
        out = self.frontend._get_homepage(json)
        self.assertEqual(out, 'foo')


class TestCPANJson(unittest.TestCase):
    def setUp(self):
        self.frontend = CPANFrontend()
        self.url = 'https://fastapi.metacpan.org/v1/release/perl-foo'
        self.version_url = 'https://fastapi.metacpan.org/v1/release/'
        self.version_url += 'AUTHOR/perl-foo-1.2'

    @requests_mock.mock()
    def test_missing_metadata(self, requests):
        requests.get(self.url, json={})
        pkg = self.frontend.parse('perl-foo')
        self.assertEqual(pkg.version, '')
        self.assertEqual(pkg.summary, '')

    @requests_mock.mock()
    def test_missing_version(self, requests):
        json = {
            'metadata': {
                'abstract': 'foo bar baz'
            }
        }
        requests.get(self.url, json=json)
        pkg = self.frontend.parse('perl-foo')
        self.assertEqual(pkg.version, '')

    @requests_mock.mock()
    def test_missing_abstract(self, requests):
        json = {
        }
        requests.get(self.url, json=json)
        pkg = self.frontend.parse('perl-foo')
        self.assertEqual(pkg.summary, '')

    @requests_mock.mock()
    def test_missing_license(self, requests):
        requests.get(self.url, json={})
        pkg = self.frontend.parse('perl-foo')
        self.assertEqual(pkg.licenses, [])

    @requests_mock.mock()
    def test_given_version(self, requests):
        fake_json = {
            'author': 'AUTHOR',
        }
        fake_version_json = {
            'release': {
                'metadata': {
                }
            }
        }
        requests.get(self.url, json=fake_json)
        requests.get(self.version_url, json=fake_version_json)
        json = self.frontend._get_json_data('perl-foo', '1.2')
        self.assertEqual(json, {'metadata': {}})

    @requests_mock.mock()
    def test_given_invalid_version(self, requests):
        fake_json = {
            'author': 'AUTHOR',
        }
        requests.get(self.url, json=fake_json)
        requests.get(self.version_url, status_code=404)
        with self.assertRaises(upt.InvalidPackageNameError):
            self.frontend._get_json_data('perl-foo', '1.2')


class TestLicenses(unittest.TestCase):
    def setUp(self):
        self.frontend = CPANFrontend()

    def test_no_licenses(self):
        self.assertEqual(self.frontend._get_licenses([]), [])

    def test_one_license(self):
        out = self.frontend._get_licenses(['bsd'])
        expected = [upt.licenses.BSDThreeClauseLicense()]
        self.assertListEqual(out, expected)

        out = self.frontend._get_licenses(['whatever'])
        expected = [upt.licenses.UnknownLicense()]
        self.assertListEqual(out, expected)

    def test_multiple_licenses(self):
        out = self.frontend._get_licenses(['freebsd', 'whatever'])
        expected = [upt.licenses.BSDTwoClauseLicense(),
                    upt.licenses.UnknownLicense()]
        self.assertListEqual(out, expected)


class TestRequirements(unittest.TestCase):
    def setUp(self):
        self.frontend = CPANFrontend()

    def test_requirements_no_prereqs(self):
        expected = {}
        out = self.frontend._get_requirements({})
        self.assertEqual(out, expected)

    def test_requirements_carp_assert_more(self):
        # Real example from
        # https://fastapi.metacpan.org/v1/release/Carp-Assert-More
        prereqs = {
            "runtime": {
                "requires": {
                    "Test::Exception": "0",
                    "Carp": "0",
                    "Carp::Assert": "0",
                    "Test::More": "0.18",
                    "Scalar::Util": "0"
                }
            },
            "configure": {
                "requires": {
                    "ExtUtils::MakeMaker": "0"
                }
            },
            "build": {
                "requires": {
                    "ExtUtils::MakeMaker": "0"
                }
            }
        }
        expected = {
            'run': [
                upt.PackageRequirement('Test::Exception', ''),
                upt.PackageRequirement('Carp', ''),
                upt.PackageRequirement('Carp::Assert', ''),
                upt.PackageRequirement('Test::More', '>=0.18'),
                upt.PackageRequirement('Scalar::Util', ''),
            ],
            'config': [
                upt.PackageRequirement('ExtUtils::MakeMaker', '')
            ],
            'build': [
                upt.PackageRequirement('ExtUtils::MakeMaker', '')
            ],
        }
        out = self.frontend._get_requirements(prereqs)
        self.assertEqual(out, expected)

    def test_requirements_appconfig(self):
        # Real example from
        # https://fastapi.metacpan.org/v1/release/AppConfig
        prereqs = {
            "build": {
                "requires": {
                    "ExtUtils::MakeMaker": "0"
                }
            },
            "test": {
                "requires": {
                    "Test::Pod": "1.0"
                }
            },
            "runtime": {
                "requires": {
                    "perl": "5.008008",
                    "Test::More": "0"
                }
            },
            "configure": {
                "requires": {
                    "ExtUtils::MakeMaker": "0"
                }
            }
        }
        expected = {
            'build': [
                upt.PackageRequirement('ExtUtils::MakeMaker', ''),
            ],
            'config': [
                upt.PackageRequirement('ExtUtils::MakeMaker', ''),
            ],
            'run': [
                upt.PackageRequirement('Test::More', ''),
            ],
            'test': [
                upt.PackageRequirement('Test::Pod', '>=1.0'),
            ],
        }
        out = self.frontend._get_requirements(prereqs)
        self.assertEqual(out, expected)


class TestArchives(unittest.TestCase):
    def setUp(self):
        self.frontend = CPANFrontend()
        self.json = {
            'download_url': 'http://www.example.com/source.tar.gz',
            'stat': {
                'size': 123,
            }
        }

    def test_get_archives(self):
        out = self.frontend._get_archives(self.json)
        self.assertEqual(len(out), 1)
        self.assertEqual(out[0].url, 'http://www.example.com/source.tar.gz')
        self.assertEqual(out[0].size, 123)


if __name__ == '__main__':
    unittest.main()
